from typing import Literal, Mapping, TypeAlias, Iterable, Sequence
import chess_notation as cn
from .sans import legal_sans, sort_key

SpecialToken: TypeAlias = Literal['[PAD]', '[CLS]', '[SEP]', '[MASK]']
PAD: TypeAlias = Literal['[PAD]']
MASK: TypeAlias = Literal['[MASK]']
Vocabulary: TypeAlias = Mapping[str | SpecialToken, int]

def remove_symbols(san: str) -> str:
  """Remove check and mate symbols from a SAN move"""
  return san.replace('+', '').replace('#', '')

def make_vocab(words: Iterable[str]) -> Vocabulary:
  return { word: i for i, word in enumerate(words) }

def legal(
  with_symbols: bool = False, *,
  pre_tokens: Sequence[str] = ['[PAD]'],
  post_tokens: Sequence[str] = [],
) -> Vocabulary:
  """Vocabulary containing all legal SAN moves
  - `with_symbols`: whether to include `+` and `#` (triples the size of the vocabulary, though)
  """
  return make_vocab(list(pre_tokens) + legal_sans(with_symbols) + list(post_tokens))

def invert(vocab: Vocabulary) -> Mapping[int, str]:
  return { i: word for word, i in vocab.items() }

default_motions = cn.MotionStyles(castles=[], pawn_captures=cn.UNIQ_PAWN_CAPTURES, piece_captures=cn.UNIQ_PIECE_CAPTURES)

def legal_styled(
  with_symbols: bool = False, *,
  motions: cn.MotionStyles = default_motions,
  effects: cn.KingEffectStyles = cn.KingEffectStyles(),
  languages: Sequence[cn.Language] = cn.LANGUAGES,
  special_tokens: Sequence[str] = ['[PAD]', '[UNK]'],
) -> Vocabulary:
  """All legal sans, styled. Size 60953 with default parameters
  - Legal SANs come first, then all representations
  """
  legal = legal_sans(with_symbols)
  words = set()
  for san in legal:
    words |= cn.all_representations(san, motions=motions, effects=effects, languages=languages)
  new_words = words - set(legal)
  sorted_words = sorted(new_words, key=sort_key)
  return make_vocab(list(special_tokens) + legal + sorted_words)
  

def extend_styled(vocab: Vocabulary, *, with_symbols: bool = False):
  """Add 'dxe' mapping to 'de' and 'de4' mapping to 'dxe4'"""
  new_vocab = dict(vocab).copy()
  for word in legal_sans(with_symbols):
    if cn.styles.is_pawn_capture(word):
      de = cn.styles.pawn_capture(word, 'de')
      dxe = cn.styles.pawn_capture(word, 'dxe')
      de4 = cn.styles.pawn_capture(word, 'de4')
      dxe4 = cn.styles.pawn_capture(word, 'dxe4')
      new_vocab[dxe] = new_vocab[de]
      new_vocab[de4] = new_vocab[dxe4]

  return new_vocab


from .models import ChessBert, ChessGPT2, ChessT5, ChessRNN
from ._loss import segmented_loss
from ._decode import segmented_argmax
from ._masking import random_masking, first_masks
from .metrics import consecutive_eq, elementwise_eq, accuracy, intra_accuracy, \
  pre_bound_accuracy, post_bound_accuracy
from . import chars, uci, fen, multi

__all__ = [
  'ChessBert', 'ChessGPT2', 'ChessT5', 'ChessRNN',
  'segmented_loss',
  'segmented_argmax', 'random_masking', 'first_masks', 'pre_bound_accuracy',
  'consecutive_eq', 'elementwise_eq', 'accuracy', 'intra_accuracy', 'post_bound_accuracy',
  'chars', 'uci', 'fen', 'multi',
]


from jaxtyping import Int
from typing import Mapping
from torch import nn, Tensor
from transformers import GPT2Config, GPT2Model

class ChessGPT2(nn.Module):
  def __init__(
    self, *, vocab_size: int, max_len: int = 512,
    hidden_size: int = 768, attention_heads: int = 12,
    pad_token_id: int, heads: Mapping[str, int] = { 'uci': 36, 'fen': 64*13 }
  ):
    super().__init__()
    self.config = GPT2Config(
      vocab_size=vocab_size,
      max_position_embeddings=max_len,
      hidden_size=hidden_size,
      num_attention_heads=attention_heads,
      pad_token_id=pad_token_id
    )
    self.gpt = GPT2Model(self.config)
    self.heads = nn.ModuleDict({
      head: nn.Linear(hidden_size, num_classes)
      for head, num_classes in heads.items()
    })

  def forward(
    self, input_ids: Int[Tensor, 'batch maxlen'],
    *, attention_mask: Int[Tensor, 'batch maxlen'] | None = None
  ):
    # Get BERT outputs
    outputs = self.gpt(input_ids=input_ids, attention_mask=attention_mask)
    sequence_output = outputs.last_hidden_state  # Shape: (batch_size, seq_len, hidden_size)
    return { head: self.heads[head](sequence_output) for head in self.heads }


from typing import Sequence, TypeAlias, Iterable
from haskellian import iter as I

piece_symbols = [None, 'P', 'N', 'B', 'R', 'Q', 'K', 'p', 'n', 'b', 'r', 'q', 'k']
symbol_ids = { symbol: i for i, symbol in enumerate(piece_symbols) }

Label: TypeAlias = Sequence[int]

def parse_fen(board_fen: str) -> Label:
  output = []
  for row in board_fen.split(' ')[0].split('/'):
    for sq in row:
      if sq.isdigit():
        output.extend([0] * int(sq))
      else:
        output.append(symbol_ids[sq])
  return output

def decode_symbols(encoded_pos: Sequence[int]) -> Sequence[Sequence[str | None]]:
  return [
    [piece_symbols[code] for code in row]
    for row in I.batch(8, encoded_pos)
  ]

def merge_empties(row: Sequence[str | None]) -> Iterable[str]:
  spaces = 0
  for piece in row:
    if piece is None:
      spaces += 1
    else:
      if spaces > 0:
        yield str(spaces)
        spaces = 0
      yield piece
  if spaces > 0:
    yield str(spaces)

def decode_fen(pred: Sequence[int]) -> str:
  symbols = decode_symbols(pred)
  return '/'.join(
    ''.join(merge_empties(row))
    for row in symbols
  )

def viz_fen(pred: Sequence[int]) -> str:
  symbols = decode_symbols(pred)
  return '/'.join(
    ''.join((c or '.') for c in row)
    for row in symbols
  )

def labels(fens: Iterable[str]) -> Sequence[Label]:
  return [parse_fen(fen) for fen in fens]


from .vocab import legal, legal_styled, remove_symbols, SpecialToken, MASK, PAD, \
  Vocabulary, invert, extend_styled

__all__ = [
  'SpecialToken', 'MASK', 'PAD', 'Vocabulary', 'extend_styled',
  'legal', 'remove_symbols', 'legal_styled', 'invert',
]


from .samples import sample_from, Sample, Batch, collate_fn
from .decode import argmax_ucis, greedy_pgn, argmax_preds
from ._metrics import loss, metrics, pt_metrics

__all__ = [
  'sample_from', 'Sample', 'Batch', 'collate_fn', 'pt_metrics',
  'argmax_ucis', 'greedy_pgn', 'loss', 'metrics', 'argmax_preds',
]


from typing import TypedDict, NamedTuple, Sequence
from haskellian import iter as I
from jaxtyping import Int
import torch
from ...multi import Sample as WordSample

class Sample(NamedTuple):
  input_ids: Int[torch.Tensor, 'L']
  fens: Int[torch.Tensor, 'B L 64']
  ucis: Int[torch.Tensor, 'B L 5']

class Batch(TypedDict):
  input_ids: Int[torch.Tensor, 'B L']
  fens: Int[torch.Tensor, 'B L 64']
  ucis: Int[torch.Tensor, 'B L 5']
  attn_mask: Int[torch.Tensor, 'B L']

def sample_from(sample: WordSample) -> Sample:
  return Sample(
    input_ids=torch.tensor(sample['input_ids']),
    fens=torch.tensor(sample['fens']),
    ucis=torch.tensor(sample['ucis'])
  )

def collate_fn(batch: Sequence[Sample], *, pad_token_id: int, ignore_idx: int = -100, maxlen: int) -> Batch:

  input_ids, fens, ucis = I.unzip(batch)
  batch_size = len(input_ids)
  maxlen = min(max(len(x) for x in input_ids), maxlen)
  padded_input_ids = torch.full((batch_size, maxlen), fill_value=pad_token_id)
  padded_fens = torch.full((batch_size, maxlen, 64), fill_value=ignore_idx)
  padded_ucis = torch.full((batch_size, maxlen, 5), fill_value=ignore_idx)
  attn_mask = torch.zeros((batch_size, maxlen))

  for i in range(batch_size):
    l = len(input_ids[i])
    padded_input_ids[i, :l] = input_ids[i]
    padded_fens[i, :l] = fens[i]
    padded_ucis[i, :l] = ucis[i]
    attn_mask[i, :l] = 1

  return Batch(input_ids=padded_input_ids, fens=padded_fens, ucis=padded_ucis, attn_mask=attn_mask)

class BatchRNN(TypedDict):
  input_ids: Int[torch.Tensor, 'B L']
  fens: Int[torch.Tensor, 'B L 64']
  ucis: Int[torch.Tensor, 'B L 5']
  seq_lens: Sequence[int]

def collate_rnn(batch: Sequence[Sample], pad_token_id: int, ignore_idx: int = -100) -> BatchRNN:

  input_ids, fens, ucis = I.unzip(batch)
  batch_size = len(input_ids)
  seq_lens = [len(x) for x in input_ids]
  maxlen = max(seq_lens)
  padded_input_ids = torch.full((batch_size, maxlen), fill_value=pad_token_id)
  padded_fens = torch.full((batch_size, maxlen, 64), fill_value=ignore_idx)
  padded_ucis = torch.full((batch_size, maxlen, 5), fill_value=ignore_idx)

  for i in range(batch_size):
    l = len(input_ids[i])
    padded_input_ids[i, :l] = input_ids[i]
    padded_fens[i, :l] = fens[i]
    padded_ucis[i, :l] = ucis[i]

  return BatchRNN(input_ids=padded_input_ids, fens=padded_fens, ucis=padded_ucis, seq_lens=seq_lens)



from typing import NamedTuple, Sequence
from ..vocab import Vocabulary
from ._labels import labels, Label

class Sample(NamedTuple):
  input_ids: Sequence[int]
  labs: Sequence[Label]

def parse_sample(
  inputs: str, fens: str, *,
  vocab: Vocabulary
) -> Sample:
	clean_inputs = inputs.strip('\n').replace('+', '').replace('#', '').split(' ')
	input_ids = [vocab[inp] for inp in clean_inputs]
	labs = labels(fens.strip('\n').split(' '))
	return Sample(input_ids, labs)


from typing import Mapping, Sequence
from jaxtyping import Int
from torch import nn, Tensor
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

class ChessRNN(nn.Module):
  def __init__(
    self, *, vocab_size: int, embed_size: int = 256,
    hidden_sizes: Sequence[int] = [256, 512, 256],
    heads: Mapping[str, int] = { 'uci': 36, 'fen': 64*13 }
  ):
    super().__init__()
    self.embed = nn.Embedding(vocab_size, embed_size)
    self.bi_lstms = nn.ModuleList([
      nn.LSTM(input_size=embed_size if i == 0 else 2*hidden_sizes[i-1], hidden_size=hidden_sizes[i], bidirectional=True)
      for i in range(len(hidden_sizes))
    ])
    self.heads = nn.ModuleDict({
      head: nn.Linear(2*hidden_sizes[-1], num_classes)
      for head, num_classes in heads.items()
    })

  def forward(self, input_ids: Int[Tensor, 'batch maxlen'], seq_lens: Sequence[int]):
    x = self.embed(input_ids)
    packed = pack_padded_sequence(x, seq_lens, batch_first=True, enforce_sorted=False) # type: ignore
    for lstm in self.bi_lstms:
      packed, _ = lstm(packed)
    z = pad_packed_sequence(packed, batch_first=True)[0]
    return { k: head(z) for k, head in self.heads.items() }


from jaxtyping import Int
import torch
from torch import Tensor

def first_masks(masked_ids: Int[torch.Tensor, 'B L'], *, mask_id: int) -> Int[torch.Tensor, 'B']:
  """Indices of the first mask in each row of `masked_ids`. Returns -1's on rows without masks."""
  have_masks = (masked_ids == mask_id).any(dim=-1)
  max_idx = torch.argmax((masked_ids == mask_id).float(), dim=-1)
  max_idx[torch.where(~have_masks)] = -1
  return max_idx

def random_masking(
  batch: Int[Tensor, 'batch maxlen'],
  attn_mask: Int[Tensor, 'batch maxlen'] | None = None,
  *, mask_id: int, mask_prob: float
):
  """Mask `batch` with `mask_id` with probability `mask_prob`.
  - If provided, positions with `attn_mask == 0` are excluded from masking.
  """
  rand = torch.rand(batch.shape, device=batch.device)
  indices = (rand < mask_prob) & (attn_mask is None or attn_mask == 1)
  output = batch.clone()
  output[indices] = mask_id
  return output



import click

def highlight_errors(pred: str, label: str) -> str:
  """Returns `pred` but highlights errors in red."""
  highlighted_prediction = ""
  for lbl_char, pred_char in zip(label, pred):
      if lbl_char == pred_char:
          highlighted_prediction += pred_char
      else:
          highlighted_prediction += click.style(pred_char, fg='red')
  # Handle the case where pred is longer than label
  if len(pred) > len(label):
      highlighted_prediction += click.style(pred[len(label):], fg='red')
  return highlighted_prediction




from typing import Sequence, TypedDict, NotRequired
from jaxtyping import Float, Int
from haskellian import iter as I
import numpy as np
import torch
from torch import Tensor, nn
from .._decode import segmented_argmax
from .._loss import segmented_loss
from ... import fen
from ..metrics import consecutive_eq, accuracy, intra_accuracy, \
  pre_bound_accuracy, post_bound_accuracy

LOGITS_ENDS = [13*(i+1) for i in range(64)]

def loss(
  logits: Float[Tensor, 'batch seq_len 64*13'],
  labels: Int[Tensor, 'batch seq_len 64'], *,
  ce_loss: nn.CrossEntropyLoss = nn.CrossEntropyLoss(ignore_index=-100)
) -> Float[Tensor, '']:
  """Cross-entropy loss summed across the board squares"""
  losses = segmented_loss(logits, labels, loss=ce_loss, ends=LOGITS_ENDS)
  return sum(l for l in losses if not torch.isnan(l)) # type: ignore

def argmax_preds(logits: Float[Tensor, 'B L 64*13']) -> Float[Tensor, 'B L 64']:
  return segmented_argmax(logits, ends=LOGITS_ENDS)

def argmax_fens(logits: Float[Tensor, 'B L 64*13']) -> list[list[str]]:
  argmaxs = argmax_preds(logits)
  return I.ndmap(fen.decode_fen, argmaxs) # type: ignore

def metrics(sample_logits: Float[Tensor, 'L 64*13'], true_fens: Sequence[str]) -> fen.Metrics:
  pred_fens = argmax_fens(sample_logits)
  return fen.metrics(fen_labs=true_fens, fen_preds=pred_fens[0])

class Metrics(TypedDict):
  acc: float
  intra_acc: float
  seq_acc: float
  consecutive_correct: float
  pre_mask_acc: NotRequired[float]
  post_mask_acc: NotRequired[float]

def pt_metrics(
  y_pred: Int[Tensor, 'B L 64'], y_true: Int[Tensor, 'B L 64'],
  *, ignore_idx: int = -100, first_masks: Int[Tensor, 'B L'] | None = None
) -> Metrics:
  y_pred = y_pred.clone()
  y_pred[y_true == ignore_idx] = ignore_idx
  consecutive_mean = float(np.mean([consecutive_eq(p, l) for p, l in zip(y_pred, y_true)]))
  
  out = Metrics(
    acc=accuracy(y_pred, y_true).item(),
    intra_acc=intra_accuracy(y_pred, y_true).item(),
    seq_acc=consecutive_mean / y_true.size(1),
    consecutive_correct=consecutive_mean,
  )
  
  if first_masks is not None:
    pre_acc = pre_bound_accuracy(y_pred, y_true, bounds=first_masks)
    if pre_acc is not None:
      out['pre_mask_acc'] = pre_acc.item()
    post_acc =  post_bound_accuracy(y_pred, y_true, bounds=first_masks)
    if post_acc is not None:
      out['post_mask_acc'] = post_acc.item()

  return out


from typing import Iterable, Sequence, TypeVar
from itertools import zip_longest

A = TypeVar('A')

def accuracy(*, preds: Iterable[A], labs: Sequence[A]) -> float:
  return sum(1 for lab, pred in zip(labs, preds) if lab == pred) / len(labs)

def seq_accuracy(*, preds: Iterable[Sequence[A]], labs: Sequence[Sequence[A]]) -> float:
  for i, (lab, pred) in enumerate(zip_longest(labs, preds)):
    if lab != pred:
      return i / len(labs)
  return 1.0

def correct_count(*, preds: Iterable[A], labs: Iterable[A]) -> int:
  count = 0
  for p, l in zip(preds, labs):
    if p != l:
      break
    count += 1
  return count


from typing import Sequence
from jaxtyping import Float, Int
import torch
import chess_utils as cu
from ...uci._labels import decode_uci

def argmax_preds(logits: Float[torch.Tensor, 'B L 36']) -> Int[torch.Tensor, 'B L 5']:
  batch_size = logits.size(0)
  from_files = torch.argmax(logits[:, :, 0:8].reshape(batch_size, -1, 8), dim=-1)
  from_ranks = torch.argmax(logits[:, :, 8:16].reshape(batch_size, -1, 8), dim=-1)
  to_files = torch.argmax(logits[:, :, 16:24].reshape(batch_size, -1, 8), dim=-1)
  to_ranks = torch.argmax(logits[:, :, 24:32].reshape(batch_size, -1, 8), dim=-1)
  proms = torch.argmax(logits[:, :, 32:36].reshape(batch_size, -1, 4), dim=-1)
  return torch.stack([from_files, from_ranks, to_files, to_ranks, proms], dim=-1)

def argmax_ucis(logits: Float[torch.Tensor, 'B L 36']) -> Sequence[Sequence[str]]:
  batch_size = logits.size(0)
  from_files, from_ranks, to_files, to_ranks, proms = argmax_preds(logits).unbind(-1)
  
  return [
    [
      decode_uci(*[int(i.item()) for i in idxs])
      for idxs in zip(from_files[b], from_ranks[b], to_files[b], to_ranks[b], proms[b])
    ]
    for b in range(batch_size)
  ]

def greedy_pgn(logits: Float[torch.Tensor, 'B L 36']) -> Sequence[Sequence[str]]:
  return [list(cu.ucis2sans(ucis)) for ucis in argmax_ucis(logits)]


from typing import Sequence, TypedDict, NotRequired
from jaxtyping import Float, Int
from torch import Tensor, nn
import torch
import numpy as np
from .._loss import segmented_loss
from ... import uci
from .decode import argmax_ucis
from ..metrics import accuracy, consecutive_eq, pre_bound_accuracy, post_bound_accuracy

def loss(
  logits: Float[Tensor, 'batch seq_len 36'],
  labels: Int[Tensor, 'batch seq_len 5'], *,
  ce_loss: nn.CrossEntropyLoss = nn.CrossEntropyLoss(ignore_index=-100)
) -> Float[Tensor, '']:
  losses = segmented_loss(logits, labels, loss=ce_loss, ends=[8, 16, 24, 32, 36])
  return sum(l for l in losses if not torch.isnan(l)) # type: ignore
  # nan may happen if all promotion labels are -100 (which is quite likely)

def metrics(sample_logits: Float[Tensor, 'L 5'], true_ucis: Sequence[str]) -> uci.Metrics:
  pred_ucis = argmax_ucis(sample_logits)
  return uci.metrics(uci_labs=true_ucis, uci_preds=pred_ucis[0])

class BatchMetrics(TypedDict):
  acc: float
  seq_acc: float
  consecutive_correct: float
  from_sq_acc: float
  to_sq_acc: float
  pre_mask_acc: NotRequired[float]
  post_mask_acc: NotRequired[float]

def pt_metrics(
  y_pred: Int[Tensor, 'B L 5'], y_true: Int[Tensor, 'B L 5'],
  *, ignore_idx: int = -100, first_masks: Int[Tensor, 'B'] | None = None
) -> BatchMetrics:
  y_pred = y_pred.clone()
  y_pred[y_true == ignore_idx] = ignore_idx
  consecutive_mean = float(np.mean([consecutive_eq(p, l) for p, l in zip(y_pred, y_true)]))
  
  out = BatchMetrics(
    acc=accuracy(y_pred, y_true).item(),
    seq_acc=consecutive_mean / y_true.size(1),
    consecutive_correct=consecutive_mean,
    from_sq_acc=accuracy(y_pred[:, :, :2], y_true[:, :, :2]).item(),
    to_sq_acc=accuracy(y_pred[:, :, 2:4], y_true[:, :, 2:4]).item(),
  )
  if first_masks is not None:
    pre_acc = pre_bound_accuracy(y_pred, y_true, bounds=first_masks)
    if pre_acc is not None:
      out['pre_mask_acc'] = pre_acc.item()
    post_acc =  post_bound_accuracy(y_pred, y_true, bounds=first_masks)
    if post_acc is not None:
      out['post_mask_acc'] = post_acc.item()

  return out


from typing import Sequence, Mapping
from jaxtyping import Float, Int
from torch import Tensor
from .. import uci, fen

def metrics(
  uci_logits: Float[Tensor, 'L 36'],
  fen_logits: Float[Tensor, 'L 64*13'],
  *,
  true_ucis: Sequence[str],
  true_fens: Sequence[str],
) -> Mapping[str, float]:
  return {
    f'uci-{k}': v for k, v in uci.metrics(uci_logits, true_ucis).items()
  } | {
    f'fen-{k}': v for k, v in fen.metrics(fen_logits, true_fens).items()
  } # type: ignore

def pt_metrics(
  uci_logits: Float[Tensor, 'B L 36'],
  fen_logits: Float[Tensor, 'B L 64*13'],
  *,
  true_ucis: Int[Tensor, 'B L 5'],
  true_fens: Int[Tensor, 'B L 64'],
  ignore_idx: int = -100,
  first_masks: Int[Tensor, 'B L'] | None = None
) -> Mapping[str, float]:
  uci_preds = uci.argmax_preds(uci_logits)
  fen_preds = fen.argmax_preds(fen_logits)
  return {
    f'uci-{k}': v for k, v in uci.pt_metrics(uci_preds, true_ucis, ignore_idx=ignore_idx, first_masks=first_masks).items()
  } | {
    f'fen-{k}': v for k, v in fen.pt_metrics(fen_preds, true_fens, ignore_idx=ignore_idx, first_masks=first_masks).items()
  } # type: ignore


from typing import Iterable
import chess

def decode_ucis(ucis: Iterable[str]) -> Iterable[str]:
  """Parses UCIs that may always have promotion to SANs. Ignores promotion if not needed"""
  board = chess.Board()
  for uci in ucis:
    try:
      move = chess.Move.from_uci(uci[:4])
      if not board.is_legal(move):
        move = chess.Move.from_uci(uci)
        if not board.is_legal(move):
          return
        
      yield board.san(move)
      board.push(move)

    except chess.InvalidMoveError:
      return


from jaxtyping import Int
from torch import nn, Tensor
from transformers import T5Config, T5Model

class ChessT5(nn.Module):
  def __init__(
    self, *, vocab_size: int, max_len: int = 512,
    hidden_size: int = 256, attention_heads: int = 4,
    pad_token_id: int = 0
  ):
    super().__init__()
    # BERT configuration
    self.config = T5Config(
      vocab_size=vocab_size,
      max_position_embeddings=max_len,
      hidden_size=hidden_size,
      num_attention_heads=attention_heads,
      pad_token_id=pad_token_id
    )
    self.t5 = T5Model(self.config)
    self.fc = nn.Linear(hidden_size, 4 * 8 + 4)

  def forward(self, input_ids: Int[Tensor, 'batch maxlen']):
    # Get BERT outputs
    outputs = self.t5(input_ids=input_ids, decoder_input_ids=input_ids)
    sequence_output = outputs.last_hidden_state  # Shape: (batch_size, seq_len, hidden_size)
    return self.fc(sequence_output)  # Shape: (batch_size, seq_len, 4 * 8 + 4 = 36)


from .samples import parse_sample, Sample
from ._decode import decode_ucis
from ._labels import labels, Label, decode_uci, decode
from ._metrics import accuracy, batch_metrics, metrics, Metrics

__all__ = [
  'parse_sample', 'Sample',
  'decode_ucis', 'labels', 'Label', 'decode',
  'accuracy', 'decode_uci', 'batch_metrics', 'metrics', 'Metrics',
]


from .samples import sample_from, Sample, Batch, collate_fn
from ._metrics import loss, argmax_fens, metrics, argmax_preds, pt_metrics

__all__ = [
  'sample_from', 'Sample', 'Batch', 'collate_fn', 'pt_metrics',
  'loss', 'argmax_fens', 'metrics', 'argmax_preds',
]


from jaxtyping import Float, Int
from torch import Tensor
import torch

def elementwise_eq(x: Float[Tensor, '*B L N'], y: Float[Tensor, '*B L N']) -> Float[Tensor, '*B L']:
  """Elementwise equality (by last dimension)"""
  return (x == y).all(dim=-1)

def accuracy(y_pred: Float[Tensor, '*B L N'], y_true: Float[Tensor, '*B L N']) -> Float[Tensor, 'L']:
  """Proportion of equal elements, averaged across samples"""
  return torch.mean(elementwise_eq(y_pred, y_true).float())

def pre_bound_accuracy(
  y_pred: Float[Tensor, '*B L N'], y_true: Float[Tensor, '*B L N'],
  *, bounds: Int[Tensor, '*B']
) -> Float[Tensor, ''] | None:
  """Accuracy within bounds. Returns None if no elements exist within bounds
  - Negative bounds indicate no bound
  """
  accs = []
  for p, l, b in zip(y_pred, y_true, bounds):
    bound = b if b >= 0 else None
    acc = accuracy(p[:bound], l[:bound])
    if not torch.isnan(acc):
      accs.append(acc)
  return torch.mean(torch.stack(accs)) if accs else None

def post_bound_accuracy(
  y_pred: Float[Tensor, '*B L N'], y_true: Float[Tensor, '*B L N'],
  *, bounds: Int[Tensor, '*B']
) -> Float[Tensor, ''] | None:
  """Accuracy after bounds. Returns None if no elements exist after bounds
  - Negative bounds ignore the whole row
  """
  accs = []
  for p, l, b in zip(y_pred, y_true, bounds):
    if b >= 0:
      acc = accuracy(p[b:], l[b:])
      if not torch.isnan(acc):
        accs.append(acc)
  return torch.mean(torch.stack(accs)) if accs else None

def intra_accuracy(y_pred: Float[Tensor, '*B L N'], y_true: Float[Tensor, '*B L N']) -> Float[Tensor, '']:
  """Average accuracy within sequence elements (dimension N)"""
  return torch.mean((y_pred == y_true).float())

def consecutive_eq(y_pred: Float[Tensor, 'L *N'], y_true: Float[Tensor, 'L *N']) -> int:
  """Number of consecutive equal elements"""
  eq = elementwise_eq(y_pred, y_true)
  i_min = eq.int().argmin()
  return y_pred.size(0) if bool(torch.min(eq[i_min])) else i_min.item() # type: ignore
  # if all elements are equal (minimum is True), return the number of elements